package org.freecompany.xmlwriter;

import java.io.*;
import java.util.*;
import javax.xml.namespace.*;
import org.freecompany.util.xml.dom2sax.Dom2Sax;
import org.xml.sax.*;
import org.xml.sax.ext.*;
import org.xml.sax.helpers.*;
import org.w3c.dom.*;

/**
 * A {@link Writer} for XML documents.
 *
 * In concept, an <code>XmlWriter</code> is a SAX {@link ContentHandler}
 * that simply writes the corresponding events as an XML document.
 */
public class XmlWriter extends FilterWriter implements ContentHandler, LexicalHandler {

	/**
	 * A value for the autoflush threshold that disables the feature.
	 */
	public static final int NO_AUTOFLUSH = -1;
	protected static final String XMLNS = "http://www.w3.org/2000/xmlns/";

	protected final Map< String, CharSequence> prefixes = new LinkedHashMap< String, CharSequence>();
	protected final Map< String, CharSequence> pending = new LinkedHashMap< String, CharSequence>();
	protected Locator locator;
	protected CharSequence version;
	protected CharSequence encoding;
	protected Boolean standalone;
	protected boolean indented;
	protected int threshold = NO_AUTOFLUSH;
	
	/**
	 * State flags for internal use.
	 */
	protected boolean startElementPending;
	protected boolean endElementGenerated;
	protected int depth;

	/**
	 * Creates an <code>XmlWriter</code> that wraps another writer.
	 * @param out the wrapped writer
	 */
	public XmlWriter( final Writer out) {
		super( out);
	}

	/**
	 * Adds a namespace mapping to the current namespace context. The
	 * next start tag written to the output will include the specified
	 * namespace prefix mapping attribute.
	 * @param uri the URI of the new namespace
	 * @param prefix the prefix used when writing tags in the namespace
	 */
	public void startPrefixMapping( String prefix, String uri) {
		prefixes.put( uri.toString(), prefix);
		pending.put( uri.toString(), prefix);
	}

	/**
	 * Adds a set of namespace mappings to the current namespace context.
	 * @param map a map of namespace URIs to prefixes.
	 */
	public void endPrefixMapping( final String prefix) {
		prefixes.remove( prefix);
	}

	/**
	 * Sets the document depth at which the underlying writer is
	 * automatically flushed.
	 * @param threshold the new threshold value
	 */
	public void setAutoFlush( final int threshold) {
		this.threshold = threshold;
	}

	/**
	 * Gets the autoflush threshold.
	 * @return the current autoflush threshold
	 */
	public int getAutoFlush() {
		return threshold;
	}

	/**
	 * Gets the current output tree depth.
	 */
	public int getDepth() {
		return depth;
	}

	/**
	 * Start an element with no attributes.
	 * @param qname the qualified name of the element
	 * @throws IOException when the underlying writer does
	 */
	public void startElement( final QName qname) throws SAXException {
		startElement( qname.getNamespaceURI(), qname.getLocalPart(), "", null);
	}

	/**
	 * Start an element with the specified attributes.
	 * @param qname the qualified name of the element
	 * @param attrs the attribute names mapped to their corresponding values
	 * @throws IOException when the underlying writer does
	 */
	public void startElement( final QName qname, final Attributes attrs) throws SAXException {
		startElement( qname.getNamespaceURI(), qname.getLocalPart(), "", attrs);
	}

	/**
	 * Start an element with no attributes.
	 * @param uri the namespace URI of the element
	 * @param name the local name of the element
	 * @throws IOException when the underlying writer does
	 */
	public void startElement( final String uri, final String name) throws SAXException {
		startElement( uri, name, "");
	}

	/**
	 * Start an element with no attributes.
	 * @param uri the namespace URI of the element
	 * @param name the local name of the element
	 * @throws IOException when the underlying writer does
	 */
	public void startElement( final String uri, final String name, final String qname) throws SAXException {
		startElement( uri, name, qname, null);
	}

	/**
	 * Start an element with the specified attributes.
	 * @param uri the namespace URI of the element
	 * @param name the local name of the element
	 * @param attrs the attribute names mapped to their corresponding values
	 * @throws IOException when the underlying writer does
	 */
	public void startElement( final String uri, final String name, final String qname, final Attributes attrs) throws SAXException {
		finishPendingStartElement();
		if ( indented && depth > 0) indent( depth);
		depth++;
		appendInternal( '<');
		writeQName( uri, "".equals( name) ? qname : name);
		if ( attrs != null) writeAttributes( attrs);
		for ( Iterator i = pending.keySet().iterator(); i.hasNext(); i.remove()) {
		   	CharSequence namespace = ( CharSequence) i.next();
			writeNamespace( namespace, pending.get( namespace));
		}
		startElementPending = true;
		endElementGenerated = false;
	}

	protected void indent( int depth) throws SAXException {
		appendInternal( '\n');
		while ( depth-- > 0) appendInternal( '\t');
	}

	/**
	 * Ends an element.
	 * @param qname the qualified name of the element
 	 * @throws IOException when the underlying writer does
	 */
	public void endElement( final QName qname) throws SAXException {
		endElement( qname.getNamespaceURI(), qname.getLocalPart(), "");
	}

	/**
	 * Ends an element.
	 * @param uri the namespace URI of the element
	 * @param name the local name of the element
	 * @throws IOException when the underlying writer does
	 */
	public void endElement( final String uri, final String name) throws SAXException {
		endElement( uri, name, "");
	}

	/**
	 * Ends an element.
	 * @param uri the namespace URI of the element
	 * @param name the local name of the element
	 * @throws IOException when the underlying writer does
	 */
	public void endElement( final String uri, final String name, final String qname) throws SAXException {
		if ( startElementPending) appendInternal( "/>");
		else {
			if ( indented && endElementGenerated) indent( depth - 1);
			appendInternal( "</");
			writeQName( uri, "".equals( name) ? qname : name);
			appendInternal( '>');
		}
		startElementPending = false;
		endElementGenerated = true;
		if ( --depth <= threshold) flushInternal();
	}

	protected void finishPendingStartElement() throws SAXException {
		if ( !startElementPending) return;
		appendInternal( '>');
		startElementPending = false;
	}

	/**
	 * Creates an empty element with no attributes.
	 * @param qname the qualified name of the element
	 * @throws IOException when the underlying writer does
	 */
	public void createElement( final QName qname) throws SAXException {
		createElement( qname.getNamespaceURI(), qname.getLocalPart(), null, null);
	}

	/**
	 * Creates an element with no attributes and the specified text content.
	 * @param qname the qualified name of the element
	 * @param content the text value of the element
	 * @throws IOException when the underlying writer does
	 */
	public void createElement( final QName qname, final CharSequence content) throws SAXException {
		createElement( qname.getNamespaceURI(), qname.getLocalPart(), null, content);
	}

	/**
	 * Creates an empty element with the specified attributes.
	 * @param qname the qualified name of the element
	 * @param attrs the attribute names mapped to their corresponding values
	 * @throws IOException when the underlying writer does
	 */
	public void createElement( final QName qname, final Attributes attrs) throws SAXException {
		createElement( qname.getNamespaceURI(), qname.getLocalPart(), attrs, null);
	}

	/**
	 * Creates an element with the specified attributes and text content.
	 * @param qname the qualified name of the element
	 * @param attrs the attribute names mapped to their corresponding values
	 * @param content the text value of the element
	 * @throws IOException when the underlying writer does
	 */
	public void createElement( final QName qname, final Attributes attrs, final CharSequence content) throws SAXException {
		createElement( qname.getNamespaceURI(), qname.getLocalPart(), attrs, content);
	}

	/**
	 * Creates an empty element with no attributes.
	 * @param uri the namespace URI of the element
	 * @param name the local name of the element
	 * @throws IOException when the underlying writer does
	 */
	public void createElement( final CharSequence uri, final CharSequence name) throws SAXException {
		createElement( uri, name, null, null);
	}

	/**
	 * Creates an element with no attributes and the specified text content.
	 * @param uri the namespace URI of the element
	 * @param name the local name of the element
	 * @param content the text value of the element
	 * @throws IOException when the underlying writer does
	 */
	public void createElement( final CharSequence uri, final CharSequence name, final CharSequence content) throws SAXException {
		createElement( uri, name, null, content);
	}

	/**
	 * Creates an empty element with the specified attributes.
	 * @param uri the namespace URI of the element
	 * @param name the local name of the element
	 * @param attrs the attribute names mapped to their corresponding values
	 * @throws IOException when the underlying writer does
	 */
	public void createElement( final CharSequence uri, final CharSequence name, final Attributes attrs) throws SAXException {
		createElement( uri, name, attrs, null);
	}

	/**
	 * Creates an element with the specified attributes and text content.
	 * @param uri the namespace URI of the element
	 * @param name the local name of the element
	 * @param attrs the attribute names mapped to their corresponding values
	 * @param content the text value of the element
	 * @throws IOException when the underlying writer does
	 */
	public void createElement( final CharSequence uri, final CharSequence name, final Attributes attrs, final CharSequence content) throws SAXException {
		startElement( uri.toString(), name.toString(), null, attrs);
		if ( content != null) characters( content);
		endElement( uri.toString(), name.toString(), null);
	}

	protected void writeQName( final CharSequence uri, final CharSequence name) throws SAXException {
		final CharSequence prefix = prefixes.get( uri);
		if ( prefix != null && !prefix.toString().equals( "")) {
			appendInternal( prefix);
			appendInternal( ':');
		}
		appendInternal( name);
	}

	protected void writeNamespace( final CharSequence uri, final CharSequence prefix) throws SAXException {
		appendInternal( ' ').appendInternal( "xmlns");
		if ( !prefix.toString().equals( "")) appendInternal( ':').appendInternal( prefix);
		appendInternal( "='").appendInternal( uri).appendInternal( '\'');
	}

	protected void writeAttributes( final Attributes attrs) throws SAXException {
		final int length = attrs.getLength();
		for ( int x = 0; x < length; x++) {
			appendInternal( ' ');
			appendInternal( attrs.getLocalName( x));
			appendInternal( "='");
			writeWithEscapes( attrs.getValue( x), "<&'"); // per XML 1.0, grammar rule 10
			appendInternal( '\'');
		}
	}

	protected void writeWithEscapes( final CharSequence value, final CharSequence charsToEscape) throws SAXException {
		for ( StringTokenizer st = new StringTokenizer( value.toString(), charsToEscape.toString(), true); st.hasMoreTokens(); ) {
			String s = st.nextToken();
			switch ( s.charAt( 0)) {
			case '\'':
				appendInternal( "&apos;");
				break;
			case '"':
				appendInternal( "&quot;");
				break;
			case '&':
				appendInternal( "&amp;");
				break;
			case '<':
				appendInternal( "&lt;");
				break;
			case '>':
				appendInternal( "&gt;");
				break;
			default:
				appendInternal( s);
			}
		}
	}

	/**
	 * Writes a DOM subtree.
	 * @param root the root of the DOM subtree
	 * @throws IOException when the underlying writer does
	 */
	public XmlWriter append( final Node root) throws SAXException {
		final Dom2Sax convertor = new Dom2Sax();
		convertor.setContentHandler( this);
		convertor.convert( root);
		return this;
	}

	protected void flushInternal() throws SAXException {
		finishPendingStartElement();
		try {
			super.flush();
		} catch ( IOException e) {
			throw new SAXException( e);
		}
	}

	/**
	 * Starts a CDATA section in the output. The caller is responsible
	 * for ensuring that the content of the CDATA section does not contain
	 * the string "]]&gt;"
	 * @throws IOException when the underlying writer does
	 */
	public void startCDATA() throws SAXException {
		finishPendingStartElement();
		appendInternal( "<![CDATA[");
		endElementGenerated = false;
	}

	/**
	 * Ends a CDATA section in the output.
	 * @throws IOException when the underlying writer does
	 */
	public void endCDATA() throws SAXException {
		appendInternal( "]]>");
		endElementGenerated = false;
	}

	/**
	 * Creates a CDATA section in the output.
	 * @throws IOException when the underlying writer does
	 */
	public void cData( CharSequence s) throws SAXException {
		startCDATA();
		appendInternal( s);
		endCDATA();
		endElementGenerated = false;
	}

	public void comment( final char[] characters, int offset, int index) throws SAXException {
		finishPendingStartElement();
		appendInternal( "<!-- ");
		appendInternal( new String( characters, offset, index));
		appendInternal( " -->");
		endElementGenerated = false;
	}

	public void startDocument() throws SAXException {
		if ( version != null || encoding != null || ( standalone != null && standalone)) {
			appendInternal( "<?xml version='");
			appendInternal( version == null ? "1.0" : version);
			if ( encoding != null) {
				appendInternal( "' encoding='");
				appendInternal( encoding);
			}
			if ( standalone != null) {
				appendInternal( "' standalone='");
				appendInternal( standalone.booleanValue() ? "yes" : "no");
			}
			appendInternal( "'?>");
			appendInternal( '\n');
		}
	}

	public void endDocument() {
	}

	public void setDocumentLocator( final Locator locator) {
		this.locator = locator;
	}

	public void ignorableWhitespace( final char[] characters, int offset, int length) throws SAXException {
		characters( characters, offset, length);
	}

	public void processingInstruction( final String target, final String data) throws SAXException {
		finishPendingStartElement();
		appendInternal( "<?").appendInternal( target).appendInternal( " ").appendInternal( data).appendInternal( "?>");
		endElementGenerated = false;
	}

	public void skippedEntity( final String name) {
		assert false;
	}

	public void startEntity( final String name) {
		assert false;
	}

	public void endEntity( final String name) {
		assert false;
	}

	public void startDTD( final String name, final String publicid, final String systemid) {
		assert false;
	}

	public void endDTD() {
		assert false;
	}

	/**
	 * Writes text to the output.
	 * @throws IOException when the underlying writer does
	 */
	public void characters( final char[] characters, int offset, int length) throws SAXException {
		characters( new String( characters, offset, length));
	}

	/**
	 * Writes character data while performing escapes as defined in the XML
	 * 1.0 specification, section 2.4.
	 */
	public void characters( final CharSequence s) throws SAXException {
		finishPendingStartElement();
		writeWithEscapes( s, "&<");
		endElementGenerated = false;
	}

	public void setVersion( final CharSequence version) {
		this.version = version;
	}

	public CharSequence getVersion() {
		return version;
	}

	public void setEncoding( final CharSequence encoding) {
		this.encoding = encoding;
	}

	public CharSequence getEncoding() {
		return encoding;
	}

	public void setStandalone( final Boolean standalone) {
		this.standalone = standalone;
	}

	public Boolean getStandalone() {
		return standalone;
	}

	public void setIndented( final boolean indented) {
		this.indented = indented;
	}

	public boolean isIndented() {
		return indented;
	}

	/**
	 * Override to convert exception types to SAX.
	 */
	public XmlWriter appendInternal( final char value) throws SAXException {
		try {
			super.append( value);
			return this;
		} catch ( IOException e) {
			throw new SAXException( e);
		}
	}

	/**
	 * Override to convert exception types to SAX.
	 */
	public XmlWriter appendInternal( final CharSequence value) throws SAXException {
		try {
			super.append( value);
			return this;
		} catch ( IOException e) {
			throw new SAXException( e);
		}
	}
}
